/*
 * Diagnostics.cpp
 *
 *  Created on: 2012-01-11
 *      Author: janez
 */

#include <string>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>

#include "Simulation.h"
#include "Diagnostics.h"
#include "ParticleList.h"
#include "Species.h"

using namespace std;

Diagnostics::Diagnostics() {
	Init();
}

Diagnostics::~Diagnostics() {
}


/*
 * Initialize object
 */
void Diagnostics::Init() {
	simulation = NULL;
	output_directory = "./";
	file = "";
	write_data_by = 0;
	n_start = n_step = n_end = next_write_n = 0;
	t_start = t_step = t_end = next_write_t = 0.0;
	n_current = 0;
	t_current = 0;
	num_files = 0;
	next_write_init = false;
	write_file_index = 0;
	write_append = header_written = false;
	species.clear();
}

/*
 * Define steps for writing data to output files. Start step, number of steps to skip and end step.
 */
void Diagnostics::set_write_data_by_step( long in_n_start, long in_n_step, long in_n_end ) {
	n_start = in_n_start;
	n_step = in_n_step;
	n_end = in_n_end;
	write_data_by = WRITE_DATA_BY_STEP;
}

/*
 * Define times for writing data to output files. Start time, time to skip and end time.
 */
void Diagnostics::set_write_data_by_time( double in_t_start, double in_t_step, double in_t_end ) {
	t_start = in_t_start;
	t_step = in_t_step;
	t_end = in_t_end;
	write_data_by = WRITE_DATA_BY_TIME;
}

/*
 * Return steps for writing data in form of vector: start, step, end.
 */
std::vector<long> Diagnostics::get_write_data_by_step() {
	std::vector<long> data;

	data.push_back(n_start);
	data.push_back(n_step);
	data.push_back(n_end);

	return data;
}

/*
 * Return times for writing data in form of vector: start, step, end.
 */
std::vector<double> Diagnostics::get_write_data_by_time() {
	std::vector<double> data;

	data.push_back(t_start);
	data.push_back(t_step);
	data.push_back(t_end);

	return data;
}

/*
 * Check if output directory exist and if so requested, create it.
 */
bool Diagnostics::checkOutputDirectoy(bool create) {
	bool allOK = false;
	struct stat buf ;


	if( stat( output_directory.c_str(), &buf ) == 0 ) {
		// output directory exists; check if it is really a directory
		if( S_ISDIR(buf.st_mode) ) {
			// OK is directory
			allOK = true;
		} else {
			// exists but not a directory
			cout << "'" << output_directory << "' exists, but is not a directory. No files will be created." << endl;
		}

	} else {
		// no, directory does not exist
		if( create ) {
			// create directory
			if( mkdir(output_directory.c_str(), 0777 ) == 0 ) {
				// all OK
				cout << "Output directory '" << output_directory << "' created." << endl;
				allOK = true;
			} else {
				// error creating directory
				cout << "Error creating output directory '" << output_directory << "'." << endl;
				allOK = false;
			}
		} else {
			cout << "Output directory '" << output_directory << "' does not exist. No files will be created." << endl;
			allOK = false;
		}
	}

	return allOK;
}

/**
 * Print content of diagnostics in human readable format.
 */
void Diagnostics::dump(int level) {
	unsigned int i;
	string offset = "";
	string disp_step_type[3] = { "UNKNOWN (error)!!", "computational step(s)", "computational time(s)" };
	string disp_type[3] = { "UNKNOWN (error)!!", "phase-space", "potential profile" };

	for( i=0 ; i < (unsigned) level ; i++ ) {
		offset += "  ";
	}

	cout << endl << offset << "Diagnostics:" << endl;
	cout << offset << "  Output directory: " << output_directory << endl;
	cout << offset << "  Output file name: " << file << endl;
	cout << offset << "  Diagnostics of: ";

	if( DIAGNOSTICS_PHASE_SPACE <= type && type <= DIAGNOSTICS_POTENTIAL_PROFILE ) {
		cout << disp_type[type] << endl;
	} else {
		cout << disp_type[0] << endl;
	}

	cout << offset << "  Species included (indexes): ";

	if( species.empty() ) {
		// all species
		cout << "all species" << endl;
	} else {
		// list all indexes...
		string sep = "";

		for( i=0 ; i < species.size() ; i++ ) {
			cout << sep << species[i];
			sep = ", ";
		}

		cout << endl;
	}

	cout << offset << "  Write points: " << endl;

	if( write_data_by == WRITE_DATA_BY_STEP ) {
		// step
		cout << offset << "   - defined are by: " << disp_step_type[WRITE_DATA_BY_STEP] << endl;
		cout << offset << "   - start step: " << n_start << endl;
		cout << offset << "   - step:       " << n_step << endl;
		cout << offset << "   - end step:   " << n_end;
		cout << ((n_end == 0 && num_files != 1) ? " (last computational step)" : "") << endl;

	} else if( write_data_by == WRITE_DATA_BY_TIME ) {
		// time
		cout << offset << "   - defined are by: " << disp_step_type[WRITE_DATA_BY_TIME] << endl;
		cout << offset << "   - start time: " << t_start << endl;
		cout << offset << "   - step time:  " << t_step << endl;
		cout << offset << "   - end time:   " << t_end;
		cout << ((t_end == 0 && num_files != 1) ? " (last computational time)" : "") << endl;
	}

	cout << offset << "   - number of files: " << num_files;
	cout << ((num_files == 0) ? " (number of files will be calculated from start and end values)" : "") << endl;
}




/**
 * Check and save some data - if conditions are OK.
 */
int Diagnostics::doWriting(){

	switch( write_data_by ) {
	case WRITE_DATA_BY_STEP:
		// check step
		return (n_current == next_write_n);

	case WRITE_DATA_BY_TIME:
		// check time
		// return true if difference between given time and next write time is
		// less that half of simulation time step
		return ( fabs(t_current - next_write_t) <= 0.5*sim_dt) ;

	}

	return 0;
}


/**
 * Compute next step/time to do writing.
 */
void Diagnostics::computeNextWritePoint() {
	long int n_max = (n_end == 0) ? maxSimStep-1 : n_end;
	double t_max = (t_end == 0) ? sim_dt * (maxSimStep-1) : t_end;

	switch( write_data_by ) {
	case WRITE_DATA_BY_STEP:
		// check step
		if( num_files == 1 ) {
			// doing only one write; define step past the max value so we will not write anything more
			next_write_n = n_max + 2;

		} else if( n_step > 0 && next_write_n < n_max ) {
			// compute next step for writing with the use of step
			next_write_n += n_step;
			//cout << "n_max = " << n_max << " next_write = " << next_write_n << endl;

		} else if( num_files  > 0 && write_file_index < num_files-1 ){
			// compute using number of files
			double new_step_diff = (write_file_index+1)*(1.0 * n_max - n_start)/(num_files-1.);
			next_write_n = (int) ( n_start + new_step_diff + 0.5 );
			//cout << "n_max = " << n_max << " next_write = " << next_write_n << endl;

		} else {
			// we write all files
			next_write_n = maxSimStep + 1;
		}

		break;

	case WRITE_DATA_BY_TIME:
		// check time
		if( num_files == 1 ) {
			// doing only one write; define step past the max value so we will not write anything more
			next_write_t = t_max + 2 * sim_dt;

		} else if( t_step > 0 && next_write_t < t_max  ) {
			// compute next step for writing with the use of time step
			next_write_t += t_step;
			//cout << "t_max = " << t_max << " next_write = " << next_write_t << endl;

		} else if( num_files  > 0 && write_file_index < num_files-1 ){
			// compute using number of files
			next_write_t = t_start + (write_file_index+1)*(1.0 * t_max - t_start)/(num_files-1.) ;
			//cout << "t_max = " << t_max << " next_write = " << next_write_t << endl;

		} else {
			// we write all files
			next_write_t = sim_dt * (maxSimStep+1);
		}

		break;
	}
}

/**
 * Check and save some data - if conditions are OK.
 */
void Diagnostics::save( long int in_n, double in_t ){
	struct stat buf ;
	FILE *fp;
	std::string file_mode = "w";


	// time and step for which we are doing saving...
	n_current = in_n;
	t_current = in_t;

	if( !next_write_init ) {
		// initialization not done yet; do it now
		if( write_data_by == WRITE_DATA_BY_STEP ) {
			next_write_n = n_start;
		} else if( write_data_by == WRITE_DATA_BY_TIME ) {
			next_write_t = t_start;
		}

		next_write_init = true;
	}

	if( !doWriting() ) {
		// nothing to do; return
		return;
	}

	// do some saving here
	//cout << " i=" << write_file_index << ", n=" << in_n << ", t=" << in_t << ": we have to save data." << endl;

	// compute next write point for current diagnostics
	computeNextWritePoint();

	if( !checkOutputDirectoy( true ) ) {
		// check for output directory; error
		return;
	}

	// create file name with output directory
	string outname = output_directory;

	if( output_directory.substr(output_directory.length()-1,1) != "/" ) {
		// add slash
		outname += "/";
	}

	if( file.find( "%" ) ) {
		// filename has printf format characters; include current file number
		char tmpfn[128];
		// sprintf( tmpfn, file.c_str(), write_file_index );
		// use current simulation step in filename
		sprintf( tmpfn, file.c_str(), n_current );
		outname += tmpfn;
	} else {
		// use given filename
		outname += file;
	}

	//cout << " - output filename: " << outname << endl;

	if( stat( outname.c_str(), &buf) == 0 && !(write_append && header_written) ) {
		// delete target file if one exists
		if( S_ISREG(buf.st_mode) ) {
			// regular file; delete it
			unlink(outname.c_str());
		} else {
			// target filename exists; display error, return
			cout << "Output filename '" << outname << "' already exists and is not a file. Writing skipped in this step." << endl;
			// increase write file index
			write_file_index ++;

			return;
		}
	}

	if( write_append ) {
		// appending to file
		file_mode = "a";
	}


	if( (fp = fopen( outname.c_str(), file_mode.c_str()) ) != NULL ) {
		// file prepared and open for writing -> now we can finally write data to file...

		if( simulation != NULL ) {
			// access to simulation data is defined;
			switch( type ) {
			case DIAGNOSTICS_PHASE_SPACE:
				// write phase space
				writePhaseSpace( fp );
				break;

			case DIAGNOSTICS_POTENTIAL_PROFILE:
				// write potential profile
				writePotentialProfile( fp );
				break;

			case DIAGNOSTICS_NUMBER_OF_PARTICLES:
				// write number of particles
				writeNumberOfParticles( fp );
				break;

			case DIAGNOSTICS_MEMORY_USAGE:
				// write current memory usage
				writeMemoryUsage( fp );
				break;

			case DIAGNOSTICS_DENSITY:
				// write density
				writeDensity( fp );
				break;

			case DIAGNOSTICS_CURRENT:
				// write density
				writeCurrent( fp );
				break;
			}
		} else {
			// display output
			cout << "ERROR: no access to simulation data (simulation pointer is NULL) - internal error. Contact program author." << endl;
		}

		// close file after writing
		fclose( fp );
	}

	// increase write file index
	write_file_index ++;
}


/**
 * Write phase-space data to given file.
 */
void Diagnostics::writePhaseSpace( FILE *fp ) {
	unsigned int i;
	Particle *pt;
	char format[] = "%d %d %g %g %g %g %g %g\n";


	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, phase-space data; row data: i species_index x y z vx vy vz\n");
		header_written = true;
	}

	if( species.size() == 0 ) {
		// write particles for all species; no sorting by species, just list of particles
		for( i=0 ; i < simulation->getParticleList()->size() ; i++ ) {
			// write data for all particles
			pt = simulation->getParticleList()->getParticle(i);
			fprintf( fp, format, i, pt->get_sp_index(),
						(pt->get_r()).e1(), pt->get_r().e2(), pt->get_r().e3(),
						pt->get_v().e1(), pt->get_v().e2(), pt->get_v().e3()
					);
		}

	} else {
		// write particles only for given species
		for( unsigned int j=0, pt_i=0 ; j < species.size() ; j++ ) {
			// print all species in list
			for( i=0 ; i < simulation->getParticleList()->size() ; i++ ) {
				// write data for all particles
				pt = simulation->getParticleList()->getParticle(i);

				if( pt->get_sp_index() == species[j] ) {
					// OK, particle of correct species; display it
					fprintf( fp, format, pt_i, pt->get_sp_index(),
								pt->get_r().e1(), pt->get_r().e2(), pt->get_r().e3(),
								pt->get_v().e1(), pt->get_v().e2(), pt->get_v().e3()
							);
					pt_i++;
				}
			}
		}

	}

}

/**
 * Write potential profile to given file.
 */
void Diagnostics::writePotentialProfile( FILE *fp ){
	unsigned i;
	double x;


	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, potential profile; row data: cell_i cell_x potential\n");
		header_written = true;
	}

	for( i=0 ; i < simulation->getPotential()->size() ; i++ ) {
		// display all potential values
		x = simulation->getSpatialRegion()->get_grid().get_x(0,i);
		fprintf(fp, "%d %g %g\n", i, x, simulation->getPotential()->at(i));
	}
}


/**
 * Write total number of particles to given file.
 */
void Diagnostics::writeNumberOfParticles( FILE *fp ){
	unsigned int i;
	Particle *pt;
	std::vector<long> num;
	std::vector<int> disp_sp;
	long total_num = 0;


	// initialize vectors...
	disp_sp.clear();
	num.clear();

	for( i=0 ; i < simulation->getSpecies()->size() ; i ++ ) {
		// initialize number of particles for each species
		num.push_back(0L);
	}

	for( i=0 ; i < simulation->getParticleList()->size() ; i++ ) {
		// now count number of particles for all species...
		pt = simulation->getParticleList()->getParticle(i);

		// increase number of particles of one species and total number of particles
		num[pt->get_sp_index()] ++;
		total_num ++;
	}

	//
	// create list of species indexes to print into file
	//
	if( species.size() == 0 ) {
		// all species
		for( i=0 ; i < simulation->getSpecies()->size() ; i ++ ) {
			disp_sp.push_back(i);
		}
	} else {
		// only selected ones...
		for( i=0 ; i < species.size() ; i ++ ) {
			disp_sp.push_back(species[i]);
		}
	}

	//
	// now write all to file...
	//
	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, number of particles; row data: i total_num");

		for( i=0 ; i < disp_sp.size() ; i ++ ) {
			fprintf( fp, " num_%s", simulation->getSpecies()->at(disp_sp[i]).get_name().c_str() );
		}

		fprintf( fp, "\n" ) ;
		header_written = true;
	}

	// now print all data
	fprintf( fp, "%ld %ld", n_current, total_num );

	for( i=0 ; i < disp_sp.size() ; i ++ ) {
		fprintf( fp, " %ld", num[disp_sp[i]] );
	}

	fprintf( fp, "\n" ) ;
}

/**
 * Write current memory usage to given file.
 */
void Diagnostics::writeMemoryUsage( FILE *fp ){
	std::string out;

	//
	// now write all to file...
	//
	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, current memory usage\n");
		header_written = true;
	}

	// get string with current memory usage
	out = simulation->getMemoryUsage(true);

//	cout << "Memory:" << endl;
//	cout << out << endl;

	// now print all data
	fprintf( fp, "%s\n", out.c_str() );
}


/**
 * Write particle (species) density to given file.
 */
void Diagnostics::writeDensity( FILE *fp ){
	unsigned int i, numCells;
	int cell_i;
	Particle *pt;
	std::vector<long> cell;
	std::vector<int> disp_sp;
	long total_num = 0;


	// initialize vectors...
	disp_sp.clear();
	numCells = simulation->getSpatialRegion()->get_grid().get_nx();
	cell.reserve(numCells);

	for( i=0 ; i < numCells ; i ++ ) {
		// initialize number of particles in each cell
		cell.push_back(0L);
	}

	for( i=0 ; i < simulation->getSpecies()->size() ; i ++ ) {
		// initial setting is to display no species
		disp_sp.push_back(0);
	}

	//
	// create list of species indexes to print into file
	//
	if( species.size() == 0 ) {
		// all species
		for( i=0 ; i < simulation->getSpecies()->size() ; i ++ ) {
			// display all species
			disp_sp[i] = 1;
		}
	} else {
		// only selected ones...
		for( i=0 ; i < species.size() ; i ++ ) {
			// display only given species
			disp_sp[species[i]] = 1;
		}
	}

	for( i=0 ; i < simulation->getParticleList()->size() ; i++ ) {
		// now count number of particles in each grid cell;
		pt = simulation->getParticleList()->getParticle(i);

		if( disp_sp[pt->get_sp_index()] == 1 ) {
			// current particle is of species that we have to count and write to file
			cell_i = simulation->getSpatialRegion()->get_grid().get_cell(0, pt->get_r().e1());

			if( cell_i > 0 ) {
				// cell is OK; increase number of particles in cell
				cell[cell_i] ++;
				total_num ++;
			}
		}
	}

	//
	// now write all to file...
	//
	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, particle density; row data: cell_index num_in_cell\n");
		header_written = true;
	}

	for( i=0 ; i < numCells ; i ++ ) {
		// now print all data
		fprintf( fp, "%d %ld\n", i, cell[i] );
	}
}


/**
 * Write particle current
 */
void Diagnostics::writeCurrent( FILE *fp ){
	//
	// now write all to file...
	//
	if( !write_append || (write_append && !header_written) ) {
		// write header of print out
		fprintf( fp, "# LTC, particle current (BETA, always species 1!!!!); row data: step, number_left_out number_right_out\n");
		header_written = true;
	}

	double q = simulation->getSpecies()->at(1).get_q();

	fprintf( fp, "%ld %g %g\n", n_current, q * simulation->getNumParticlesLeftOut(), q * simulation->getNumParticlesRightOut() );
}

