/*
 * Diagnostics.h
 *
 *  Created on: 2012-01-11
 *      Author: janez
 */

#ifndef DIAGNOSTICS_H_
#define DIAGNOSTICS_H_

#include <string>
#include <vector>

#include "peg_tc.h"
#include "ParticleList.h"
#include "Species.h"
#include "SpatialRegion.h"

#define WRITE_DATA_BY_STEP		1
#define WRITE_DATA_BY_TIME		2

class Simulation;

class Diagnostics {
	Simulation* simulation;				// pointer to simulation data
	std::string output_directory;		// name of output directory
	std::string file;					// file name; can include printf format characters
	long n_current;
	double t_current;
	bool next_write_init;
	int write_data_by;					// write data according to time or computing step
	long next_write_n;					// next step when to do writing
	double next_write_t;				// next time when to do writing
	long n_start, n_step, n_end;		// start, step and end computational step for writing data to file
	double t_start, t_step, t_end;		// start, step and end time for writing data to file
	int type;							// diagnostics type: potential profile, phase-space, etc.
	long num_files;						// number of files to write
	std::vector<int> species;			// list of species to include; if list is empty, include all species in system
	long int maxSimStep;				// used when number of files is defined and no ending step
	double sim_dt;						// dt in simulation -- used in condition for writing to file
	long write_file_index;				// sequence of next file to be written and how many files were already written
	bool write_append;					// appending to file or overwriting file content
	bool header_written;				// is header (first line in file) already written or not -- used when appending to file

	int doWriting();
	bool checkOutputDirectoy(bool create);
	void computeNextWritePoint();
	void writePhaseSpace( FILE *fp );
	void writePotentialProfile( FILE *fp );
	void writeNumberOfParticles( FILE *fp );
	void writeMemoryUsage( FILE *fp );
	void writeDensity( FILE *fp );
	void writeCurrent( FILE *fp );

public:
	static const int DIAGNOSTICS_PHASE_SPACE = 1;
	static const int DIAGNOSTICS_POTENTIAL_PROFILE = 2;
	static const int DIAGNOSTICS_NUMBER_OF_PARTICLES = 3;
	static const int DIAGNOSTICS_MEMORY_USAGE = 4;
	static const int DIAGNOSTICS_DENSITY = 5;
	static const int DIAGNOSTICS_CURRENT = 6;

	Diagnostics();
	virtual ~Diagnostics();
	void Init();
	void dump(int level);

	long long memory_footprint() { return sizeof(this) + species.size()*sizeof(int); };

	inline void setSimulationDataAccess( Simulation *inSim ) { simulation = inSim; };
	inline void set_diagnostics_type( int in_type ) { type = in_type; };
	inline void set_output_directory( std::string dir ) { output_directory = dir; };
	inline void set_file( std::string in_file ) { file = in_file; };
	inline void set_number_of_files( long in_num ) { num_files = in_num; };
	inline void set_max_step( long int in_max_step ) { maxSimStep = in_max_step; };
	inline void set_dt( double in_dt ) { sim_dt = in_dt; };
	inline void set_write_data_by( int in_type ) { write_data_by = in_type; } ;
	void set_write_data_by_step( long in_n_start, long in_n_step, long in_n_end );
	inline void set_write_data_by_step_start(long in) { n_start = in; };
	inline void set_write_data_by_step_step(long in) { n_step = in; };
	inline void set_write_data_by_step_end(long in) { n_end = in; };
	void set_write_data_by_time( double in_t_start, double in_t_skip, double in_t_end );
	inline void set_write_data_by_time_start(double in) { t_start = in; };
	inline void set_write_data_by_time_step(double in) { t_step = in; };
	inline void set_write_data_by_time_end(double in) { t_end = in; };
	inline void species_add( int spec_index ) { species.push_back(spec_index); };
	inline void species_clear() { species.clear(); };
	inline void set_write_append( bool inAppend ) { write_append = inAppend; };

	inline int get_diagnostics_type() { return type; };
	inline std::string get_output_directory() { return output_directory; };
	inline std::string get_file() { return file; };
	inline long get_number_of_files() { return num_files; };
	inline int get_write_data_by() { return write_data_by; } ;
	std::vector<long> get_write_data_by_step();
	inline long get_write_data_by_step_start() { return n_start; };
	inline long get_write_data_by_step_step() { return n_step; };
	inline long get_write_data_by_step_end() { return n_end; };
	std::vector<double> get_write_data_by_time();
	inline double get_write_data_by_time_start() { return t_start; };
	inline double get_write_data_by_time_step() { return t_step; };
	inline double get_write_data_by_time_end() { return t_end; };
	inline bool get_write_append() { return write_append; };

	void save( long int in_n, double in_t );


};

#endif /* DIAGNOSTICS_H_ */
