/*
 * PlasmaDevice.cpp
 *
 *  Created on: Aug 7, 2011
 *      Author: janez
 */

#include "SpatialRegion.h"
#include "ParticleList.h"
#include <iostream>
#include <vector>

//#define DEBUG_BOUNDARY_CHECK

using namespace std;


SpatialRegion::SpatialRegion() {
	init();

}

SpatialRegion::~SpatialRegion() {

}

/**
 * Initialize spatial region
 */
void SpatialRegion::init() {
	dt = V0 = V1 = area = 0;
	p2c = 1;
	maxSimStep = 0;
	compute_method = Compute::COMPUTE_METHOD_DEFAULT;
	grid.init();
}


/**
 * Display spatial region data in human readable form.
 */
void SpatialRegion::dump(int level) {
	string offset = "";

	for( int i=0 ; i < level ; i++ ) {
		offset += "  ";
	}

	cout << endl << offset << "Spatial region:" << endl;
	cout << offset << "  Time step: " << dt << endl;
	cout << offset << "  max. simulation step: " << maxSimStep << endl;
	cout << offset << "  Boundaries: V0=" << V0 << ", V1=" << V1 << endl;
	// display grid;
	grid.dump(level+1);
}

/**
 * Check if minimal set of data is defined.
 * Return true if yes, and false of not enough data is defined.
 */
bool SpatialRegion::defined(bool dispErrors) {
	bool ret = false;
	vector<string> mess;


	if( dt > 0 ) {
		// OK, time step is defined;
		ret = true;
	}

	if( ret && IS_ZERO(area) ) {
		// region area is not defined
		mess.push_back("Spatial region cross section is not defined (parameter 'area').");
		ret = false;
	}


	if( ret && maxSimStep <= 0 ) {
		// number of simulation steps is not defined;
		mess.push_back("Number of simulation steps to run is not defined (maxSimStep).");
		ret = false;
	}

	if( ret && !grid.defined(dispErrors) ) {
		// grid is not defined; errors are displayed inside grid.defined
		ret = false;
	}

	if( dispErrors && mess.size() > 0 ) {
		// display errors
		for( unsigned int i=0 ; i < mess.size() ; i++ ) {
			cout << mess[i] << endl;
		}
	}

	// clear message(s)
	mess.clear();

	return ret;
}



/**
 * Check boundaries for their interaction with particles: absorption, reflection.
 * Method deletes particles that are stuck in boundary and change positions and
 * velocities for particles that are reflected.
 */
void SpatialRegion::checkBoundaries( ParticleList& pl ){
	unsigned int i, max_i;
	vector<unsigned int> particlesToDelete;
	double x;

	// initialize array
	particlesToDelete.clear();

	for( i=0, max_i=pl.size() ; i < max_i ; i++ ) {
		// check all particles
		x = pl.getParticle(i)->get_r().e1();

		if( x < get_grid_x_min() || x > get_grid_x_max() ) {
			// particle is out of current region; mark it for deletion
			particlesToDelete.push_back(i);
		}
	}

#ifdef DEBUG_BOUNDARY_CHECK
	cout << "Particles: " << pl.size() << ", particles to delete: " << particlesToDelete.size() << endl;
#endif

	if( particlesToDelete.size() > 0 ) {
		// there are some particles to delete
		pl.deleteParticles(particlesToDelete);
	}

#ifdef DEBUG_BOUNDARY_CHECK
	cout << "Particles: " << pl.size() << endl;
#endif
}


/**
 * Count number of particles outside of right-hand wall
 */
long SpatialRegion::countOutRight( ParticleList& pl ){
	unsigned int i, max_i;
	long out = 0;
	double x;

	for( i=0, max_i=pl.size() ; i < max_i ; i++ ) {
		// check all particles
		x = pl.getParticle(i)->get_r().e1();

		if( x > get_grid_x_max() ) {
			// particle is out
			out ++;
		}
	}

	return out;
}



/**
 * Count number of particles outside of left-hand wall
 */
long SpatialRegion::countOutLeft( ParticleList& pl ){
	unsigned int i, max_i;
	long out = 0;
	double x;

	for( i=0, max_i=pl.size() ; i < max_i ; i++ ) {
		// check all particles
		x = pl.getParticle(i)->get_r().e1();

		if( x < get_grid_x_max() ) {
			// particle is out
			out ++;
		}
	}

	return out;
}

