/*
 * Tools.cpp
 *
 *  Created on: Jun 24, 2012
 *      Author: janez
 */

#include "Tools.h"

#include <string>
#include <iostream>
#include <vector>

using namespace std;

Tools::Tools() {
	type = TOOLS_NOT_DEFINED;
}

Tools::~Tools() {
}

/**
 * Run selected tool.
 * @param type -- tool to run (VDF, ...)
 * @param argc -- number of arguments (number of arguments already processed are not included in this number)
 * @param argv -- parameters for tool (already processed parameters are not in given array)
 */
void Tools::run( int argc, char** argv) {

	switch( type ) {
	case TOOLS_VDF:
		// generate velocity distribution functions
		// parameters: 1= input file (full name, including directory + file + extension
		// 2= coordiate (x,y,z); 3= x; 4= dx; 5= N_vx (number of sections in velocity direction)
		// optional:
		// 6= species (index), 7= v_min, 8=v_max
		if( argc >= 5 ) {
			// OK, get parameters;
			std::string file = argv[0];
			std::string coord_str = argv[1];
			double x = string2double( argv[2] );
			double dx = string2double( argv[3] );
			int n = atoi( argv[4] );
			double v_min = 0, v_max = 0;
			int species = -1;

			if( n < 1 ) {
				// error
				printf( "Error: number of cells in velocty (current: %d) positive and non-zero.\n", n );
				return;
			}

			if( argc >= 6 ) {
				// read which species to process
				species = atoi( argv[5] );
			}

			if( argc >= 7 ) {
				// read also minimal velocity
				v_min = string2double( argv[6] );
			}

			if( argc >= 8 ) {
				// read also maximal velocity
				v_max = string2double( argv[7] );
			}

			// default is X direction
			int coord = 0;

			if( coord_str == "y" ) {
				// display Y direction (coordinate)
				coord = 1;

			} else if( coord_str == "z" ) {
				// display Z direction (coordinate)
				coord = 2;
			}

			// run creating VDF
			createVDF( file, coord, x, dx, n, v_min, v_max, species );

		} else {
			//
			printf("Error: missing parameters for VDF.\n" );
			printf("Usage:\n ltf -vdf <file> <coordinate> <x> <dx> <n_cells> [<species index>] [<v_min> <v_max>]\n\n");
			printf("- where:\n");
			printf("   <file>: file with data to read\n");
			printf("   <coordinate>: which coordinate to use: x, y oz z; default is x (values other than listed)\n");
			printf("   <x>: at which value to make DF; strip is made around this value\n");
			printf("   <dx>: width of strip\n");
			printf("   <n_cells>: number of cells in v direction\n");
			printf("   <species>: for which species to create VDF (file can have more than one species); use -1 for all species in file; optional parameter\n");
			printf("   <v_min>: value for minimal value for velocity; optional value\n");
			printf("   <v_max>: value for maximal value for velocity; optional value\n");
			printf("\nTo automatically define velocity limits, do not define values for v_min and v_max.\n");
		}

		break;


	case TOOLS_NOT_DEFINED:
	default:
		// when not defined or unknown tool defined
		printf( "Error in parameters for tools.\n");
		break;
	}
}


/**
 * Create velocity distribution function (VDF) from given file and position "in_x" and width "in_dx" and "n_cells" in velocity direction.
 * @param file     -- file from which to read data
 * @param coord    -- for which coordinate to create VDF (0=x, 1=y, 2=z)
 * @param in_x     -- position where to create VDF (position)
 * @param in_dx    -- width of strip (in_x is in the middle of the strip)
 * @param n_cells  -- number of cells in velocity direction
 * @param in_v_min -- minimal value for velocity
 * @param in_v_max -- maximal value for velocity
 * @param in_species - for which species to create VDF; -1 means all species in file
 *
 * To automatically define velocity limits (v_min and v_max), define 0 for both in_v_min and in_v_max.
 *
 */
void Tools::createVDF( std::string file, int in_coord, double in_x, double in_dx, int n_cells, double in_v_min, double in_v_max, int in_species ) {
	int tmp1, species, v_pos;
	double x_min, x_max, v_min, v_max;
	double x[3], v[3];
	char line[1024];
	FILE *fp;
	std::vector<long> values;


	if( in_coord < 0 && in_coord > 2 ) {
		// given coordinate out of range; use 0 (x coordinate)
		in_coord = 0;
	}

	if( (fp = fopen(file.c_str(), "r")) == NULL ) {
		// error opening file
		cout << "Error opening file " << file << ". End." << endl;
		return;
	}

	// get limits of parameters in file
	getMinMaxValues( fp, in_coord, in_coord, &x_min, &x_max, &v_min, &v_max );

	// initialize result values
	values.reserve( n_cells );

	for( int i=0 ; i < n_cells ; i++ ) {
		// fill with zeros
		values.push_back(0);
	}

	// limits for strip
	x_min = in_x - in_dx/2;
	x_max = x_min + in_dx;

	if( !( in_v_min == 0 && in_v_max == 0) ) {
		// velocity limits are defined; use them
		v_min = in_v_min;
		v_max = in_v_max;
	}

	while( !feof(fp) ) {
		// find first line with data
		fgets( line, 1023, fp );

		if( line[0] != '#' ) {
			// current line is not comment; it's ok
			sscanf(line, "%d %d %lf %lf %lf %lf %lf %lf", &tmp1, &species, &x[0], &x[1], &x[2], &v[0], &v[1], &v[2]);

			if( x_min <= x[in_coord] && x[in_coord] < x_max && v_min <= v[in_coord] && v[in_coord] <= v_max ) {
				/* current line is inside strip and velocity limits */
				if( in_species == -1 || species == in_species ) {
					// correct species
					/* find in which cell it falls */
					v_pos = (v[in_coord]-v_min)/(v_max-v_min)*(n_cells-1);
					/* add one count to right cell */
					values[v_pos] ++;
				}
			}
		}
	}

	// close file
	fclose(fp);

	//
	// Now display results
	//
	// first print header (as comment)...
	char coords[3] = { 'x', 'y', 'z' };

	printf("# VDF for coordinate: %c, file: %s\n", coords[in_coord], file.c_str() ) ;
	printf("# x: %g, dx: %g, v_min: %g, v_max: %g\n", in_x, in_dx, v_min, v_max);

	double dv = (v_max - v_min) / n_cells;
	double v_mid = v_min + dv / 2;

	for( unsigned int i=0 ; i < values.size() ; i++ ) {
		// ...then print all values
		printf("%d %g %ld\n", i, v_mid, values[i]);
		v_mid += dv;
	}

	printf("# end.\n");

	return;
}


/**
 * Convert string to double.
 */
double Tools::string2double(char* str) {
    char* endptr;
    double value = strtod(str, &endptr);

    if ( endptr == str || *endptr) {
    	// empty string or not valid number (string starting with non-number char)
    	value = 0;
    }

    return value;
}


/**
 * Read values from file and define "bounding" box = min and max values for x and vx
 * Coordinates start with 0 (x coord. is 0, etc); col_x (coord. number for x values), col_y (coord. number for v values);
 *
 */
void Tools::getMinMaxValues( FILE *fp, int coord_x, int coord_v, double *x1, double *x2, double *v1, double *v2 ){
	double x[3], v[3];
	int seq, species;
	bool init_done = false;
	char line[1024];


	/* go back to beginning */
	rewind( fp );

	while( !feof(fp) ) {
		// find first line with data
		fgets( line, 1023, fp );

		if( line[0] != '#' ) {
			// current line is not comment; it's ok
			sscanf( line, "%d %d %lf %lf %lf %lf %lf %lf", &seq, &species, &x[0], &x[1], &x[2], &v[0], &v[1], &v[2]);

			if( !init_done ) {
				// initialization (first good line)
				*x1 = *x2 = x[coord_x];
				*v1 = *v2 = v[coord_v];
				init_done = true;
			} else {
				// data line; check values;
				if( x[coord_x] < *x1 ) *x1 = x[coord_x];
				if( x[coord_x] > *x2 ) *x2 = x[coord_x];
				if( v[coord_v] < *v1 ) *v1 = v[coord_v];
				if( v[coord_v] > *v2 ) *v2 = v[coord_v];
			}
		}
	}

	/* go back to beginning */
	rewind( fp );
}
