/**
 * Tree.cpp
 *
 *  Created on: Aug 22, 2011
 *      Author: janez
 */

#include "TreeNode.h"

#include <string>
#include <iostream>
#include <stdlib.h>
#include <cstdio>
using namespace std;

//#define DEBUG_ALLOC

/**
 * New TreeNode without particle vector allocated.
 */
TreeNode::TreeNode() {
	init(0);
}

/**
 * New TreeNode with reserved space for given number of particles.
 */
TreeNode::TreeNode( unsigned long noParticles ) {
	init(noParticles);
}

/**
 * New TreeNode with reserved space for given number of particles and added with given particle indexes.
 */
TreeNode::TreeNode( std::vector<int> indexes ) {
	init(indexes.size());
	this->addParticleIndexs(indexes);
}



TreeNode::~TreeNode() {
	/* free array of particle indexes, if they exist */
	this->particle_index.clear();
	// then first free child nodes
	if( this->l1 != NULL ) delete this->l1;
	if( this->l2 != NULL ) delete this->l2;

}

/**
 * Initialize object
 */
void TreeNode::init( unsigned long noParticles ) {
	particle_index.clear();
	particle_index.reserve(noParticles);
	l1 = l2 = NULL;
	ds = de = Qc = Xc = 0.0;
	index = level = 0;
}


/**
 * Dump node data.
 */
void TreeNode::dump( int level ) {
	string offset = "";

	for( int i=0 ; i < level ; i++ ) {
		offset += "  ";
	}

	cout << offset << "level: " << this->level << ", node index: " << this->index;
	cout << ", particles: " << this->particle_index.size() << ", ds: " << this->ds << ", de: " << this->de;
	cout << "  Qc: " << this->Qc << ", Xc: " << this->Xc << endl;

	if( this->l1 != NULL ) {
		this->l1->dump( level+1 );
	}

	if( this->l2 ) {
		this->l2->dump( level+1 );
	}
}

/**
 * Add particle indexes in given vector to current tree node.
 */
void TreeNode::addParticleIndexs(std::vector<int> inIndexes){
	for( unsigned int i=0; i < inIndexes.size() ; i++ ) {
		/* store particle indexes into node */
		if( inIndexes[i] < 0 ) {
			cout << "particle index is less than 0. Level: " << this->level << ", num nodes: " << this->index << endl;
			exit(1);
		}

		this->particle_index.push_back(inIndexes[i]);
	}
}


/**
 * How many memory is tree currently using
 */
long long TreeNode::memory_footprint() {
	long long size = 0;

	/* size of empty node */
	size = sizeof(TreeNode);

	// size of particle list
	size += sizeof(this->particle_index);

	if( this->l1 != NULL ) {
		/* child exists, get his size */
		size += this->l1->memory_footprint();
	}

	if( this->l2 != NULL ) {
		/* child exists, get his size */
		size += this->l2->memory_footprint();
	}

	return size;
};







/**
 * Compute charge of cluster (sum of charges of particles in cluster)
 */
void TreeNode::computeQc(ParticleList *pl) {
	double Qc = 0;
	unsigned int i;


	if( this->l1 != NULL ) {
		// compute Xc in left leaf
		this->l1->computeQc(pl);
		Qc += this->l1->getQc();
	}

	if( this->l2 != NULL ) {
		// compute Xc in right leaf
		this->l2->computeQc(pl);
		Qc += this->l2->getQc();
	}

	if( this->particle_index.size() > 0 ) {
		// add also particles in current node
		for( i=0 ; i < this->particle_index.size() ; i++) {
			/* sum charges of all particles in node */
			Qc += pl->getParticle(this->particle_index[i])->get_q();
		}
	}

	// store computed value
	this->Qc = Qc;
}

/**
 * Compute center of node.
 */
void TreeNode::computeXc() {

	if( this->l1 != NULL ) {
		// compute Xc in left leaf
		this->l1->computeXc();
	}

	if( this->l2 != NULL ) {
		// compute Xc in right leaf
		this->l2->computeXc();
	}

	// compute in current node
	this->Xc = (this->ds + this->de)/2;

	return ;
}


/*
 * Count particles in node and both
 */
long TreeNode::count_particles() {
	long num = 0;

	if( this->l1 == NULL && this->l2 == NULL ) {
		// no leafs
		num = particle_index.size();
	} else {
		// check both leafs and add number of particles in each node (if exists)
		if( this->l1 != NULL ) {
			num += this->l1->count_particles();
		}

		if( this->l2 != NULL ) {
			num += this->l1->count_particles();
		}
	}

	return num;
}
